from functools import partial

import vexpr as vp
import vexpr.core as core
import vexpr.numpy as vnp
import vexpr.numpy.primitives as p
import vexpr.vectorization as v

def stack_vectorize(expr):
    # get unique list of ops, preserving order
    vexpr_ops = list(dict.fromkeys(v.op for v in expr.args[0]
                                   if isinstance(v, vp.Vexpr)))
    vexpr_ops = [op for op in vexpr_ops if op != core.symbol_p]

    if len(vexpr_ops) > 0:
        for allow_partial in (False, True):
            for op in vexpr_ops:
                try:
                    return v.pushthrough(expr, op,
                                            allow_partial=allow_partial)
                except v.CannotVectorize:
                    pass

    return expr


# TODO: obviously, understand which of these impls should be the same.
def concatenate_vectorize(expr):
    # get unique list of ops, preserving order
    vexpr_ops = list(dict.fromkeys(v.op
                                   for v in expr.args[0]
                                   if isinstance(v, vp.Vexpr)))

    vexpr_ops = [op for op in vexpr_ops if op != core.symbol_p]

    # TODO: in general, every user-provided stack (and maybe concatenate, and
    # other things?) should be pushed through before trying other stuff. they
    # just cause confusion. One solution is to do a first pass that is bottom
    # up, performing a vectorize on each stack, starting from the leaf nodes.
    # another solution is to add special logic to ~every passthrough that tries
    # to eliminate certain ops from the children before proceeding with specific
    # logic.


    # TODO if any child ops are a concatenate with the same axis, absorb their
    # children

    # TODO: for any adjacent numpy arrays, perform a concat immediately?

    if len(vexpr_ops) > 0:
        # first pass, detect any stacks and vectorize them, attempting to make them disappear.
        if p.stack_p in vexpr_ops:
            child_exprs = []
            changed = False
            for child_expr in expr.args[0]:
                if child_expr.op == p.stack_p:
                    try:
                        child_expr2 = v._vectorize(child_expr)
                        # TODO find a faster way of doing this. probably make
                        # _vectorize indicate whether anything changed, perhaps
                        # always throwing CannotVectorize if no change occurs.
                        if child_expr != child_expr2:
                            child_expr = child_expr2
                            changed = True
                    except v.CannotVectorize:
                        pass
                child_exprs.append(child_expr)
            if changed:
                expr = vnp.concatenate(child_exprs, **expr.kwargs)
                return concatenate_vectorize(expr)

        for allow_partial in (False, True):
            for op in vexpr_ops:
                try:
                    return v.pushthrough(expr, op,
                                         allow_partial=allow_partial)
                except v.CannotVectorize:
                    pass


    return expr

def reduction_vectorize(op, expr):
    assert expr.op == op
    # TODO handle axis?
    child_expr = expr.args[0]
    if not isinstance(child_expr, core.Vexpr):
        child_expr = vnp.stack(expr.args[0])
    child_expr = v._vectorize(child_expr)
    kwargs = {}
    if "axis" in expr.kwargs:
        kwargs["axis"] = expr.kwargs["axis"]
    return core.Vexpr(op, (child_expr,), kwargs)

v.vectorize_impls.update({
    p.stack_p: stack_vectorize,
    p.concatenate_p: concatenate_vectorize,
    p.sum_p: partial(reduction_vectorize, p.sum_p),
    p.prod_p: partial(reduction_vectorize, p.prod_p),
})



from vexpr.core import _p_and_constructor

# Array construction
zeros_p, zeros = _p_and_constructor("numpy.zeros")
ones_p, ones = _p_and_constructor("numpy.ones")
# Array manipulation
stack_p, stack = _p_and_constructor("numpy.stack")
concatenate_p, concatenate = _p_and_constructor("numpy.concatenate")
moveaxis_p, moveaxis = _p_and_constructor("numpy.moveaxis")
reshape_p, reshape = _p_and_constructor("numpy.reshape")
# Unary operators
exp_p, exp = _p_and_constructor("numpy.exp")
log_p, log = _p_and_constructor("numpy.log")
# Reduction operators
sum_p, sum = _p_and_constructor("numpy.sum")
prod_p, prod = _p_and_constructor("numpy.prod")
# Indexed reductions
add_at_p, add_at = _p_and_constructor("numpy.add.at")
multiply_at_p, multiply_at = _p_and_constructor("numpy.multiply.at")


__all__ = [
    "zeros",
    "ones",
    "stack",
    "concatenate",
    "moveaxis",
    "reshape",
    "exp",
    "log",
    "sum",
    "prod",
    "add_at", # TODO somehow change to "add.at"
    "multiply_at", # TODO somehow change to "multiply.at"
]



import torch

from vexpr import core
from vexpr.torch.register_eval import allow_listlike_arg0
from . import primitives as p

def shuffle_impl(arr, indices, dim=0):
    with torch.profiler.record_function("shuffle"):
        selection = [slice(None)] * arr.ndim
        selection[dim] = indices
        return arr[selection]

core.eval_impls[p.shuffle_p] = shuffle_impl



def split_and_stack_impl(x, lengths, expanded_length, expanded_indices,
                         max_length, fill_value=0., dim=0):
    """
    Technically, this function receives redundant information. It receives
    the list of lengths because this makes it easier for the vectorizer to
    concatenate calls to split_and_pad, so that it doesn't have to reconstruct
    the lengths from the expanded_indices.
    """
    with torch.profiler.record_function("split_and_stack"):
        if dim < 0:
            dim += len(x.shape)

        final_shape = (*x.shape[:dim], len(lengths),
                       max_length,
                       *x.shape[dim + 1:])

        if expanded_length == x.shape[dim]:
            # no expansion needed
            return x.view(final_shape)

        expanded_shape = (*x.shape[:dim], expanded_length,
                          *x.shape[dim + 1:])

        x_expanded = x.new_full(expanded_shape, fill_value)
        selection = [slice(None)] * len(expanded_shape)
        selection[dim] = expanded_indices
        selection = tuple(selection)
        x_expanded[selection] = x
        return x_expanded.view(final_shape)

core.eval_impls[p.split_and_stack_p] = split_and_stack_impl


def cdist_multi_impl(x1, x2, p, dim=0):
    with torch.profiler.record_function("cdist_multi"):
        f = torch.vmap(torch.cdist, in_dims=(-2, -2), out_dims=dim)
        return f(x1, x2, p=p)

core.eval_impls[p.cdist_multi_p] = cdist_multi_impl


def sum_multi_impl(x, dim=None):
    with torch.profiler.record_function("sum_multi"):
        if dim is None:
            batch_dim = 0
        elif dim < 0:
            batch_dim = dim - 1
        else:
            batch_dim = dim

        # Use vmap so that "dim" can still be in the language of the sum. For
        # example, when performing a set of sums on tensors over dim 0, without
        # vmap we would make dim 0 a batch dimension and convert the dim to 1.
        f = torch.vmap(torch.sum, in_dims=batch_dim, out_dims=dim)
        return f(x, dim=dim)

core.eval_impls[p.sum_multi_p] = sum_multi_impl


def fast_prod_positive_impl(x, dim=None, epsilon=1e-10):
    with torch.profiler.record_function("fast_prod_positive"):
        # Much faster than torch.prod in the backward pass, since it doesn't require
        # a GPU synchronize.
        # https://twitter.com/mrcslws/status/1589721597396815873
        return x.clamp(min=epsilon).log().sum(dim=dim).exp()

core.eval_impls[p.fast_prod_positive_p] = allow_listlike_arg0(
    fast_prod_positive_impl)


def fast_prod_positive_multi_impl(x, dim=None, epsilon=1e-10):
    with torch.profiler.record_function("fast_prod_positive_multi"):
        if dim is None:
            batch_dim = 0
        elif dim < 0:
            batch_dim = dim - 1
        else:
            batch_dim = dim

        # Use vmap so that "dim" can still be in the language of the prod. For
        # example, when performing a set of prods on tensors over dim 0, without
        # vmap we would make dim 0 a batch dimension and convert the dim to 1.
        f = torch.vmap(fast_prod_positive_impl, in_dims=batch_dim,
                       out_dims=dim)
        return f(x, dim=dim, epsilon=epsilon)

core.eval_impls[p.fast_prod_positive_multi_p] = fast_prod_positive_multi_impl


def prod_multi_impl(x, dim=None):
    with torch.profiler.record_function("prod_multi"):
        if dim is None:
            batch_dim = 0
        elif dim < 0:
            batch_dim = dim - 1
        else:
            batch_dim = dim

        # Use vmap so that "dim" can still be in the language of the prod. For
        # example, when performing a set of prods on tensors over dim 0, without
        # vmap we would make dim 0 a batch dimension and convert the dim to 1.
        f = torch.vmap(torch.prod, in_dims=batch_dim, out_dims=dim)
        return f(x, dim=dim)

core.eval_impls[p.prod_multi_p] = prod_multi_impl

def index_add_into_zeros_impl(n_sums, dim, index, source, *args, **kwargs):
    with torch.profiler.record_function("index_add_into_zeros"):
        shape = list(source.shape)
        shape[dim] = n_sums
        shape = tuple(shape)
        return source.new_zeros(shape).index_add_(dim, index, source, *args,
                                                  **kwargs)

core.eval_impls[p.index_add_into_zeros_p] = index_add_into_zeros_impl


def index_reduce_into_ones_impl(n_reductions, dim, index, source, *args,
                                **kwargs):
    with torch.profiler.record_function("index_reduce_into_ones"):
        shape = list(source.shape)
        shape[dim] = n_reductions
        shape = tuple(shape)
        ret = torch.ones(shape, dtype=source.dtype, device=source.device)
        return ret.index_reduce(dim, index, source, *args, **kwargs)

core.eval_impls[p.index_reduce_into_ones_p] = index_reduce_into_ones_impl


def heads_tails_impl(alpha):
    with torch.profiler.record_function("heads_tails"):
        if isinstance(alpha, torch.Tensor) and alpha.ndim > 0:
            # Build [alpha1, 1-alpha1, alpha2, 1-alpha2, ...]
            return torch.stack([alpha, 1.0 - alpha], dim=-1).view(*alpha.shape[:-1], -1)
        else:
            return torch.stack([alpha, 1.0 - alpha])

core.eval_impls[p.heads_tails_p] = heads_tails_impl



from functools import partial

import numpy as np

import vexpr as vp
import vexpr.numpy as vnp
import vexpr.core as core
import vexpr.numpy.primitives as p
import vexpr.vectorization as v

def canonical_axis(axis, ndim):
    if axis is None:
        return None
    elif axis < 0:
        return axis + ndim
    else:
        return axis

def push_concatenate_through_getitem(expr, allow_partial=True):
    assert expr.op == p.concatenate_p

    # Quick hacky assumptions: every child expr is selecting from a symbol, and
    # they are the same symbol.
    # TODO: if any child is not a getitem, then push through all the getitems,
    # concatenate with the non-getitems, and insert a shuffle after everything
    # to correct the order.
    if not all(sub_expr.op == core.operator_getitem_p
               for sub_expr in expr.args[0]):
        raise NotImplementedError()

    select_targets = [sub_expr.args[0]
                      for sub_expr in expr.args[0]]
    select_target = select_targets[0]
    if not all(target.op == core.symbol_p for target in select_targets):
        raise NotImplementedError()
    if not all(target == select_target
               for target in select_targets):
        raise NotImplementedError()

    selections = [sub_expr.args[1]
                  for sub_expr in expr.args[0]]

    all_select_axes = []
    all_select_indices = []

    for selection in selections:
        if isinstance(selection, tuple):
            # parse selection tuple to determine which axis contains a list of
            # indices
            select_axis = None
            select_indices = None
            for i, sel in enumerate(selection):
                if isinstance(sel, slice):
                    continue
                elif sel is Ellipsis:
                    break
                else:
                    if select_axis is not None:
                        raise NotImplementedError(
                            "Can't handle indexing on multiple axes yet.")
                    select_axis = i
                    select_indices = sel

            if select_axis is None:
                # an ellipsis was found, count from the back
                for i, sel in enumerate(reversed(selection)):
                    if isinstance(sel, slice):
                        continue
                    elif sel is Ellipsis:
                        if select_axis is None:
                            raise ValueError(
                                "Found multiple ellipses, or found no indices",
                                selection)
                    else:
                        if select_axis is not None:
                            raise NotImplementedError(
                                "Can't handle indexing on multiple axes yet.")
                        select_axis = -(i + 1)
                        select_indices = sel

        else:
            select_axis = 0
            select_indices = selection

        all_select_axes.append(select_axis)
        all_select_indices.append(select_indices)

    if not all(select_axis == all_select_axes[0] for select_axis in all_select_axes):
        raise NotImplementedError()
    select_axis = all_select_axes[0]

    axis = expr.kwargs.get("axis", 0)
    ndim = len(v.shape(select_target))
    if not canonical_axis(axis, ndim) == canonical_axis(select_axis, ndim):
        raise NotImplementedError()
    else:
        # prefer the version that the user specified
        axis = select_axis

    indices = np.concatenate(all_select_indices)
    if axis == 0:
        new_selection = indices
    elif axis < 0:
        new_selection = (Ellipsis, indices,) + (slice(None),) * ((-axis) - 1)
    else:
        new_selection = ((slice(None),) * axis) + (indices,)

    return select_target[new_selection]

v.pushthrough_impls.update({
    (p.concatenate_p, core.operator_getitem_p): push_concatenate_through_getitem,
})



def push_moveaxis_through_stack(expr, allow_partial=True):
    assert expr.op == p.moveaxis_p
    assert isinstance(expr.args[0], vp.Vexpr) and expr.args[0].op == p.stack_p

    source = expr.args[1]
    stack_expr = expr.args[0]
    stack_axis = stack_expr.kwargs.get("axis", 0)

    ndim = len(v.shape(stack_expr))
    if source < 0: source += ndim
    if stack_axis < 0: stack_axis += ndim

    if stack_axis == source:
        dest = expr.args[2]
        return vnp.stack(stack_expr.args[0], axis=dest)
    else:
        # don't attempt, leave the moveaxis where it is
        return expr


def push_moveaxis_through_sum(expr, allow_partial=True):
    assert expr.op == p.moveaxis_p
    assert isinstance(expr.args[0], vp.Vexpr) and expr.args[0].op == p.sum_p

    sum_expr = expr.args[0]
    sum_arg0 = sum_expr.args[0]
    if not isinstance(sum_arg0, vp.Vexpr):
        sum_arg0 = vnp.stack(sum_arg0)

    source = expr.args[1]
    dest = expr.args[2]
    sum_arg0 = vnp.moveaxis(sum_arg0, source, dest)
    sum_arg0 = v.pushthrough(sum_arg0, p.stack_p)

    sum_axis = sum_expr.kwargs.get("axis", None)

    trace = list(range(len(v.shape(sum_expr))))
    trace.insert(dest, trace.pop(source))
    new_axis = trace.index(sum_axis)

    return sum(sum_arg0, axis=new_axis)

v.pushthrough_impls.update({
    (p.moveaxis_p, p.stack_p): push_moveaxis_through_stack,
    (p.moveaxis_p, p.sum_p): push_moveaxis_through_sum,
})


def push_concatenate_through_stack(expr, allow_partial=True):
    assert expr.op == p.concatenate_p

    # initial hack: assume all args are stack_p
    assert all(isinstance(arg, vp.Vexpr) and arg.op == p.stack_p
               for arg in expr.args[0])
    assert expr.kwargs.get("axis", 0) == 0

    # (when i remove the hack, the result is going to be
    # shuffle(concatenate([stack, non-stack-exprs])), or remove the outer
    # concatenate of there are no non-stack-exprs.)

    # todo push vectorize through children
    all_stacked_vexprs = []
    for child_expr in expr.args[0]:
        all_stacked_vexprs.extend(child_expr.args[0])

    expr = vnp.stack(all_stacked_vexprs)
    expr = v._vectorize(expr)
    return expr

def np_stack_shape(initial_shape, num_elements, axis=0):
    """
    generated by github copilot
    """
    if axis < 0:
        axis += len(initial_shape) + 1
    return initial_shape[:axis] + (num_elements,) + initial_shape[axis:]

def push_stack_through_reduction(reduction_p, parallel_reduction, expr,
                                 allow_partial=True):
    assert expr.op == p.stack_p

    exprs_to_stack = expr.args[0]
    all_reduction_operands = []
    at_indices = []

    stack_axis = expr.kwargs.get("axis", 0)

    for i, child_expr in enumerate(exprs_to_stack):
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == reduction_p:

            r_axis = child_expr.kwargs.get("axis", None)
            if r_axis is None:
                raise NotImplementedError()

            r_arg0 = child_expr.args[0]
            if isinstance(r_arg0, vp.Vexpr):
                num_operands = v.shape(r_arg0)[r_axis]
            else:
                num_operands = len(r_arg0)
                # treat array_like as an implicit stack.
                r_arg0 = vnp.stack(r_arg0)

            if r_axis != stack_axis:
                prev_shape = v.shape(r_arg0)
                r_arg0 = vnp.moveaxis(r_arg0, r_axis, stack_axis)
                r_arg0 = v.pushthrough(r_arg0, child_expr.op)
                r_axis = stack_axis

            # Incorporate child_expr's computation into a vectorized reduction.
            all_reduction_operands.append(r_arg0)
            at_indices += [i] * num_operands
        else:
            # Pass child_expr through. Implement Identity as a reduction of 1
            # element.

            # TODO is this is gross to do a full vectorize here?
            child_expr = vnp.stack([child_expr], axis=stack_axis)
            child_expr = v._vectorize(child_expr)
            all_reduction_operands.append(child_expr)
            at_indices.append(i)

    at_indices = np.array(at_indices)

    if stack_axis == 0:
        pass
    elif stack_axis < 0:
        at_indices = (Ellipsis, at_indices,) + (slice(None),) * ((-stack_axis) - 1)
    else:
        at_indices = ((slice(None),) * stack_axis) + (at_indices,)

    child_shape = next(v.shape(expr) for expr in exprs_to_stack
                       if isinstance(expr, vp.Vexpr))

    result_shape = np_stack_shape(child_shape, len(exprs_to_stack),
                                  axis=stack_axis)

    return parallel_reduction(result_shape,
                              at_indices,
                              v._vectorize(vnp.concatenate(all_reduction_operands,
                                                              axis=stack_axis)))

def parallel_sum(result_shape, indices, source):
    return vnp.add_at(vnp.zeros(result_shape), indices, source)

def parallel_prod(result_shape, indices, source):
    return vnp.multiply_at(vnp.ones(result_shape), indices, source)

def push_concatenate_through_truediv(expr, allow_partial=True):
    assert expr.op == p.concatenate_p

    # initial hack: assume all args are truediv
    #
    # TODO: add similar treatment of "*_through_mul", dividing by 1.0. It's
    # easier than mul, since it's unambiguous; we can always put 1.0 in the
    # denominator.
    assert all(isinstance(arg, vp.Vexpr) and arg.op == core.operator_truediv_p
               for arg in expr.args[0])

    num = []
    den = []
    for child_expr in expr.args[0]:
        num.append(child_expr.args[0])
        den.append(child_expr.args[1])

    axis = expr.kwargs.get("axis", 0)
    num = v._vectorize(vnp.concatenate(num, axis=axis))
    den = v._vectorize(vnp.concatenate(den, axis=axis))

    return num / den

def push_stack_through_mul(expr, allow_partial=True):
    assert expr.op == p.stack_p

    left = []
    right = []
    for child_expr in expr.args[0]:
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == core.operator_mul_p:
            left.append(child_expr.args[0])
            right.append(child_expr.args[1])
        else:
            # TODO: how to choose side? for example, some users do w * matrix,
            # while others do matrix * w. Here I want to append the 1 to the w.
            # It seems the way to do this is to check which symbols each
            # child_expr is derived from, then when we're doing an identity, use
            # the symbols *this* child_expr is derived from. if they aren't
            # derived from the same symbols, then there's no benefit to choosing
            # correctly.
            #
            # TODO: what if we're multiplying matrices with matrices? sometimes
            # we need to instead append a matrix of ones, etc.
            left.append(1.0)
            right.append(child_expr)

    kwargs = {}
    if "axis" in expr.kwargs:
        kwargs["axis"] = expr.kwargs["axis"]

    # decide whether ones need to be inserted into the shape to get proper
    # broadcasting
    left_shapes = [v.shape(child_expr) for child_expr in left]
    left_shape_before_stack = left_shapes[0]
    assert all(left_shape_before_stack == shape for shape in left_shapes)
    right_shapes = [v.shape(child_expr) for child_expr in right]
    right_shape_before_stack = right_shapes[0]
    assert all(right_shape_before_stack == shape for shape in right_shapes)
    discrepancy = len(right_shape_before_stack) - len(left_shape_before_stack)
    if discrepancy == 0:
        implicit_left_shape = left_shape_before_stack
        implicit_right_shape = right_shape_before_stack
    elif discrepancy > 0:
        implicit_left_shape = left_shape_before_stack + (1,) * discrepancy
        implicit_right_shape = right_shape_before_stack
    else:
        implicit_left_shape = left_shape_before_stack
        implicit_right_shape = right_shape_before_stack + (1,) * -discrepancy

    axis = expr.kwargs.get("axis", 0)
    actual_left_shape = np_stack_shape(left_shape_before_stack,
                                       len(left), axis)
    actual_right_shape = np_stack_shape(right_shape_before_stack,
                                        len(right), axis)
    implicit_left_shape = np_stack_shape(implicit_left_shape,
                                         len(left), axis)
    implicit_right_shape = np_stack_shape(implicit_right_shape,
                                          len(right), axis)

    implicit_neg_axis = (axis
                         if axis < 0
                         else -len(implicit_left_shape) + axis)

    left = v._vectorize(vnp.stack(left, **kwargs))
    if implicit_left_shape[implicit_neg_axis:] \
       != actual_left_shape[implicit_neg_axis:]:
        left = vnp.reshape(left, implicit_left_shape[implicit_neg_axis:])

    right = v._vectorize(vnp.stack(right, **kwargs))
    if implicit_right_shape[implicit_neg_axis:] \
       != actual_right_shape[implicit_neg_axis:]:
        right = vnp.reshape(right, implicit_right_shape[implicit_neg_axis:])

    return left * right

def push_concatenate_through_mul(expr, allow_partial=True):
    assert expr.op == p.concatenate_p

    axis = expr.kwargs.get("axis", 0)

    left = []
    right = []
    for child_expr in expr.args[0]:
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == core.operator_mul_p:
            left.append(child_expr.args[0])
            right.append(child_expr.args[1])
        else:
            # TODO: how to choose side? for example, some users do w * matrix,
            # while others do matrix * w. Here I want to append the 1 to the w.
            # It seems the way to do this is to check which symbols each
            # child_expr is derived from, then when we're doing an identity, use
            # the symbols *this* child_expr is derived from. if they aren't
            # derived from the same symbols, then there's no benefit to choosing
            # correctly.
            #
            # TODO: what if we're multiplying matrices with matrices? sometimes
            # we need to instead append a matrix of ones, etc.
            num_values = v.shape(child_expr)[axis]
            left.append(np.ones(num_values))
            right.append(child_expr)

    left = v._vectorize(vnp.concatenate(left, axis=axis))
    right = v._vectorize(vnp.concatenate(right, axis=axis))

    return left * right

v.pushthrough_impls.update({
    (p.concatenate_p, p.stack_p): push_concatenate_through_stack,
    (p.stack_p, p.sum_p): partial(push_stack_through_reduction, p.sum_p,
                                  parallel_sum),
    (p.stack_p, p.prod_p): partial(push_stack_through_reduction, p.prod_p,
                                   parallel_prod),
    (p.concatenate_p, core.operator_truediv_p): push_concatenate_through_truediv,
    (p.stack_p, core.operator_mul_p): push_stack_through_mul,
    (p.concatenate_p, core.operator_mul_p): push_concatenate_through_mul,
})



from functools import partial

import torch

import vexpr as vp
import vexpr.custom.torch as vctorch
import vexpr.custom.torch.primitives as csp_p
import vexpr.torch as vtorch
import vexpr.torch.primitives as p
import vexpr.vectorization as v
from vexpr.custom.torch.utils import (
    split_and_stack_kwargs,
)
from vexpr.torch.utils import (
    torch_cat_shape,
    torch_stack_shape,
    cat_remainder_then_combine,
    push_stack_through_reduction,
)


def push_cat_through_shuffle(expr, allow_partial=True):
    assert expr.op == p.cat_p

    cat_dim = expr.kwargs.get("dim", 0)

    # get a list of shuffles with the same dim by wrapping everything else with
    # identity shuffles
    child_exprs = [(child_expr
                    if child_expr.op == csp_p.shuffle_p
                    and child_expr.kwargs.get("dim", 0) == cat_dim
                    else v.with_return_shape(vctorch.shuffle(
                            child_expr,
                            torch.arange(v.shape(child_expr)[cat_dim])),
                                             v.shape(child_expr)))
                   for child_expr in expr.args[0]]

    grandchildren = []
    indices = []
    base = 0
    for child_expr in child_exprs:
        grandchildren.append(child_expr.args[0])
        child_indices = child_expr.args[1]
        indices.append(child_indices + base)
        base += len(child_indices)

    result_shape = torch_cat_shape([v.shape(child_expr)
                                    for child_expr in child_exprs],
                                   cat_dim)

    ret = v._vectorize(
        v.with_return_shape(
            vtorch.cat(grandchildren, **expr.kwargs),
            result_shape
        )
    )
    indices = torch.cat(indices)

    if not torch.equal(indices, torch.arange(len(indices))):
        ret = v.with_return_shape(vctorch.shuffle(ret, indices, **expr.kwargs),
                                  result_shape)

    return ret


v.pushthrough_impls[(p.cat_p, csp_p.shuffle_p)] = push_cat_through_shuffle


def combine_split_and_stack(exprs, dim=0):
    assert all(expr.op == csp_p.split_and_stack_p
               for expr in exprs)

    lengths = []
    children = []
    for expr in exprs:
        assert expr.op == csp_p.split_and_stack_p
        lengths += expr.kwargs["lengths"]
        children.append(expr.args[0])

    children = v._vectorize(vtorch.cat(children, dim=dim))

    return vctorch.split_and_stack(
        children,
        **split_and_stack_kwargs(lengths),
        dim=dim,
    )


def push_cat_through_cdist_multi(expr, allow_partial=True):
    assert expr.op == p.cat_p

    metric = next(child_expr.kwargs.get("p", 2)
                  for child_expr in expr.args[0]
                  if child_expr.op == csp_p.cdist_multi_p)

    cat_dim = expr.kwargs.get("dim", 0)

    left = []
    right = []
    axes = []

    applicable_exprs = []
    applicable_indices = []
    remainder = []
    remainder_indices = []
    base = 0
    for child_expr in expr.args[0]:
        num_indices = v.shape(child_expr)[cat_dim]
        result_indices = list(range(base, base + num_indices))
        if (child_expr.op == csp_p.cdist_multi_p
            and child_expr.kwargs.get("p", 2) == metric):
            applicable_indices += result_indices
            applicable_exprs.append(child_expr)
            left.append(child_expr.args[0])
            right.append(child_expr.args[1])
            axes.append(child_expr.kwargs.get("dim", None))
        else:
            if not allow_partial:
                raise v.CannotVectorize()
            remainder_indices += result_indices
            remainder.append(child_expr)
        base += num_indices

    canonicalized_axes = [(axis if axis is not None else 0)
                          for axis in axes]
    if not all(axis == canonicalized_axes[0]
               for axis in canonicalized_axes[1:]):
        raise ValueError("Expected same axes", axes)
    axis = axes[0]

    left = combine_split_and_stack(left, dim=-1)
    right = combine_split_and_stack(right, dim=-1)

    kwargs = dict(p=metric)
    if axis is not None:
        kwargs["dim"] = axis

    return cat_remainder_then_combine(
        v.with_return_shape(
            vctorch.cdist_multi(left, right, **kwargs),
            torch_cat_shape([v.shape(child_expr)
                             for child_expr in applicable_exprs],
                            dim=cat_dim)),
        remainder,
        applicable_indices,
        remainder_indices,
        **expr.kwargs)


v.pushthrough_impls[(p.cat_p, csp_p.cdist_multi_p)] = push_cat_through_cdist_multi


def push_cat_through_reduction_multi(reduction_multi_p, parallel_reduction,
                                     fill_value, expr, all_partial=True):
    assert expr.op == p.cat_p

    cat_dim = expr.kwargs.get("dim", 0)

    reduction_dims = [child_expr.kwargs.get("dim", None)
                      for child_expr in expr.args[0]
                      if isinstance(child_expr, vp.Vexpr)
                      and child_expr.op == reduction_multi_p]
    assert all(dim == reduction_dims[0] for dim in reduction_dims)
    reduction_dim = reduction_dims[0]

    lengths = []
    grandchildren = []
    for child_expr in expr.args[0]:
        if isinstance(child_expr, vp.Vexpr) \
           and child_expr.op == reduction_multi_p:
            split_and_stack_expr = child_expr.args[0]
            assert split_and_stack_expr.op == csp_p.split_and_stack_p
            lengths += split_and_stack_expr.kwargs["lengths"]
            grandchildren.append(split_and_stack_expr.args[0])
        else:
            grandchildren.append(child_expr)
            lengths += [1] * v.shape(child_expr)[cat_dim]

    grandchildren = v._vectorize(vtorch.cat(grandchildren, dim=cat_dim))
    grandchildren = vctorch.split_and_stack(grandchildren,
                                            **split_and_stack_kwargs(lengths),
                                            fill_value=fill_value,
                                            dim=reduction_dim)

    return v.with_return_shape(parallel_reduction(grandchildren,
                                                  dim=reduction_dim),
                               torch_cat_shape([v.shape(child_expr)
                                                for child_expr in expr.args[0]],
                                               dim=cat_dim))

v.pushthrough_impls.update({
    (p.cat_p, csp_p.sum_multi_p): partial(
        push_cat_through_reduction_multi, csp_p.sum_multi_p, vctorch.sum_multi, 0.),
    (p.cat_p, csp_p.prod_multi_p): partial(
        push_cat_through_reduction_multi, csp_p.prod_multi_p, vctorch.prod_multi, 1.),
    (p.cat_p, csp_p.fast_prod_positive_multi_p): partial(
        push_cat_through_reduction_multi, csp_p.fast_prod_positive_multi_p,
        vctorch.fast_prod_positive_multi, 1.),
    (p.stack_p, csp_p.fast_prod_positive_p): partial(
        push_stack_through_reduction, csp_p.fast_prod_positive_p,
        vctorch.fast_prod_positive_multi, 1.)
})


def push_cat_through_index_reduction_into(
        index_reduction_into_p, parallel_reduction,
        expr, allow_partial=True):
    assert expr.op == p.cat_p

    cat_dim = expr.kwargs.get("dim", 0)

    index_reduction_dims = [child_expr.args[1]
                            for child_expr in expr.args[0]
                            if isinstance(child_expr, vp.Vexpr)
                            and child_expr.op == index_reduction_into_p]
    assert all(dim == index_reduction_dims[0] for dim in index_reduction_dims)
    index_reduction_dim = index_reduction_dims[0]

    indices = []
    grandchildren = []
    base = 0
    for child_expr in expr.args[0]:
        child_shape = v.shape(child_expr)
        num_results = child_shape[cat_dim]
        if isinstance(child_expr, vp.Vexpr) \
           and child_expr.op == index_reduction_into_p:
            grandchildren.append(child_expr.args[3])
            indices.append(child_expr.args[2] + base)
            base += num_results
        else:
            grandchildren.append(child_expr)
            indices.append(torch.arange(base, base + num_results))
            base += num_results

    indices = torch.cat(indices)
    grandchildren = v._vectorize(vtorch.cat(grandchildren, dim=cat_dim))
    num_sums = base
    return v.with_return_shape(parallel_reduction(num_sums, index_reduction_dim,
                                                  indices, grandchildren),
                               torch_cat_shape([v.shape(child_expr)
                                                for child_expr in
                                                expr.args[0]],
                                               dim=cat_dim))

def parallel_sum(num_sums, dim, index, source):
    return vctorch.index_add_into_zeros(num_sums, dim, index, source)

def parallel_prod(num_reductions, dim, index, source):
    return vctorch.index_reduce_into_ones(num_reductions, dim, index, source,
                                          "prod")

v.pushthrough_impls.update({
    (p.cat_p, csp_p.index_add_into_zeros_p): partial(
        push_cat_through_index_reduction_into, csp_p.index_add_into_zeros_p,
        parallel_sum),
    # TODO this is hardcoded to prod, but index reduce might use e.g. mean
    (p.cat_p, csp_p.index_reduce_into_ones_p): partial(
        push_cat_through_index_reduction_into, csp_p.index_reduce_into_ones_p,
        parallel_prod),
})


def push_concat_through_heads_tails(expr, allow_partial=True):
    assert expr.op == p.cat_p

    assert all(isinstance(child_expr, vp.Vexpr)
               and child_expr.op == csp_p.heads_tails_p
               for child_expr in expr.args[0])

    if len(expr.args[0]) == 1:
        return v._vectorize(expr.args[0][0])

    cat_dim = expr.kwargs.get("dim", 0)
    alphas = []
    for child_expr in expr.args[0]:
        if (not isinstance(child_expr, vp.Vexpr)
            or child_expr.op != csp_p.heads_tails_p):
            raise NotImplementedError()

        alpha = child_expr.args[0]
        alpha_shape = v.shape(alpha)
        if len(v.shape(alpha)) == 0:
            alpha = v.with_return_shape(
                v._vectorize(vtorch.stack([alpha], dim=cat_dim)),
                (1,))
        alphas.append(alpha)

    alphas = v._vectorize(v.with_return_shape(
        vtorch.cat(alphas, dim=cat_dim),
        torch_cat_shape([v.shape(alpha)
                         for alpha in alphas],
                        dim=cat_dim)
        ))
    return v.with_return_shape(vctorch.heads_tails(alphas),
                               torch_cat_shape([v.shape(child_expr)
                                                for child_expr in expr.args[0]],
                                               dim=cat_dim))

v.pushthrough_impls.update({
    (p.cat_p, csp_p.heads_tails_p): push_concat_through_heads_tails
})



import vexpr.custom.torch as vctorch
import vexpr.custom.torch.primitives as cp
import vexpr.vectorization as v


def heads_tails_vectorize(expr):
    return v.with_return_shape(vctorch.heads_tails(v._vectorize(expr.args[0])),
                               v.shape(expr))

v.vectorize_impls[cp.heads_tails_p] = heads_tails_vectorize



import torch

import vexpr.vectorization as v
from vexpr import core
from . import primitives as p

def shape(t):
    if isinstance(t, torch.Tensor):
        return tuple(t.shape)
    else:
        return ()

v.shape_impls.update({
    torch.Tensor: shape,
})

v.shape_impls.update({
    dtype: shape
    for dtype in [torch.float16, torch.float32, torch.float64, torch.int8,
                  torch.int16, torch.int32, torch.int64, torch.uint8,
                  torch.bool]
})

def allow_listlike_arg0(torch_func):
    def wrapper(arg0, *args, **kwargs):
        with torch.profiler.record_function("listlike_of_tensors"):
            if isinstance(arg0, (list, tuple)):
                if len(arg0) > 1 and isinstance(arg0[0], torch.Tensor):
                    return torch_func(torch.stack(arg0), *args, **kwargs)
                else:
                    return torch_func(torch.tensor(arg0), *args, **kwargs)
            else:
                return torch_func(arg0, *args, **kwargs)
    return wrapper

def listlike_of_tensors(torch_func):
    def wrapper(arg0, *args, **kwargs):
        with torch.profiler.record_function("listlike_of_tensors"):
            if isinstance(arg0, (list, tuple)):
                return torch_func([torch.as_tensor(x)
                                   for x in arg0], *args, **kwargs)
            else:
                return torch_func(arg0, *args, **kwargs)
    return wrapper

core.eval_impls.update({
    p.zeros_p: torch.zeros,
    p.ones_p: torch.ones,
    p.stack_p: listlike_of_tensors(torch.stack),
    p.cat_p: torch.cat,
    p.moveaxis_p: torch.moveaxis,
    p.reshape_p: torch.reshape,
    p.exp_p: torch.exp,
    p.log_p: torch.log,
    p.sum_p: allow_listlike_arg0(torch.sum),
    p.prod_p: allow_listlike_arg0(torch.prod),
    p.cdist_p: torch.cdist,
    p.scatter_p: torch.scatter,
})


def index_add_impl(a, dim, index, source, *args, **kwargs):
    return a.index_add(dim, index, source, *args, **kwargs)

def index_reduce_impl(a, dim, index, source, *args, **kwargs):
    return a.index_reduce(dim, index, source, *args, **kwargs)

core.eval_impls.update({
    p.index_add_p: index_add_impl,
    p.index_reduce_p: index_reduce_impl,
})



import numpy as np
import scipy

from vexpr import core
from .primitives import cdist_multi_p

def cdist_multi_impl(x1, x2, lengths, axis=0):
    base = 0
    results = []
    for length in lengths:
        results.append(scipy.spatial.distance.cdist(x1[..., base:base+length],
                                                    x2[..., base:base+length]))
        base += length
    return np.stack(results, axis=axis)

core.eval_impls[cdist_multi_p] = cdist_multi_impl



from vexpr.core import _p_and_constructor

shuffle_p, shuffle = _p_and_constructor("custom.torch.shuffle")
cdist_multi_p, cdist_multi = _p_and_constructor("custom.torch.cdist_multi")
fast_prod_positive_p, fast_prod_positive = _p_and_constructor(
    "custom.torch.fast_prod_positive")
fast_prod_positive_multi_p, fast_prod_positive_multi = _p_and_constructor(
    "custom.torch.fast_prod_positive_multi")
sum_multi_p, sum_multi = _p_and_constructor("custom.torch.sum_multi")
prod_multi_p, prod_multi = _p_and_constructor("custom.torch.prod_multi")

# calls index_add, sizing the return value according to the input value's shape
# this is like doing a vtorch.index_add(vtorch.zeros(...), ...), but the shape
# of the zeros is dynamically chosen.
index_add_into_zeros_p, index_add_into_zeros = _p_and_constructor("custom.torch.index_add_into_zeros")
index_reduce_into_ones_p, index_reduce_into_ones = _p_and_constructor("custom.torch.index_reduce_into_ones")

split_and_stack_p, split_and_stack = _p_and_constructor("custom.torch.split_and_stack")

heads_tails_p, heads_tails = _p_and_constructor("custom.torch.heads_tails")

__all__ = [
    "shuffle",
    "cdist_multi",
    "fast_prod_positive",
    "fast_prod_positive_multi",
    "sum_multi",
    "prod_multi",
    "index_add_into_zeros",
    "index_reduce_into_ones",
    "split_and_stack",
    "heads_tails",
]



import numpy as np

from vexpr import core
import vexpr.vectorization
from . import primitives as p

vexpr.vectorization.shape_impls.update({
    np.ndarray: np.shape,
})

vexpr.vectorization.shape_impls.update({
    dtype: np.shape
    for _, dtypes in np.sctypes.items()
    for dtype in dtypes
})

core.eval_impls.update({
    p.zeros_p: np.zeros,
    p.ones_p: np.ones,
    p.stack_p: np.stack,
    p.concatenate_p: np.concatenate,
    p.moveaxis_p: np.moveaxis,
    p.reshape_p: np.reshape,
    p.exp_p: np.exp,
    p.log_p: np.log,
    p.sum_p: np.sum,
    p.prod_p: np.prod,
})

def add_at_impl(a, indices, b):
    np.add.at(a, indices, b)
    return a
def multiply_at_impl(a, indices, b):
    np.multiply.at(a, indices, b)
    return a
core.eval_impls.update({
    p.add_at_p: add_at_impl,
    p.multiply_at_p: multiply_at_impl,
})



import numpy as np

import vexpr.numpy as vnp
import vexpr.numpy.primitives as np_p
import vexpr.custom.scipy.primitives as csp_p
import vexpr.vectorization as v
from vexpr import core
from vexpr.custom.scipy import cdist_multi

from . import primitives as p

def push_stack_through_cdist(expr, allow_partial=True):
    assert expr.op == np_p.stack_p
    assert all(child_expr.op == p.cdist_p for child_expr in expr.args[0])

    # TODO process this
    stack_axis = expr.kwargs.get("axis", 0)

    left = []
    right = []
    lengths = []
    child_matrix_shapes = []
    for child_expr in expr.args[0]:
        left.append(child_expr.args[0])
        right.append(child_expr.args[1])

        shape = v.shape(child_expr.args[0])
        length = shape[-1]
        lengths.append(length)
        child_matrix_shapes.append(shape[:-1])

    left = v._vectorize(vnp.concatenate(left, axis=-1))
    right = v._vectorize(vnp.concatenate(right, axis=-1))

    kwargs = dict(
        lengths=np.array(lengths)
    )
    if "axis" in expr.kwargs:
        kwargs["axis"] = expr.kwargs["axis"]

    ret = cdist_multi(left, right, **kwargs)

    # Compute shape of result
    child_matrix_shape = child_matrix_shapes[0]
    assert all(shape == child_matrix_shape for shape in child_matrix_shapes[1:])

    axis = expr.kwargs.get("axis", 0)
    # use axis to determine result shape after stack
    result_shape = child_matrix_shape
    if axis < 0:
        axis += len(result_shape) + 1
    result_shape = result_shape[:axis] + (len(lengths),) + result_shape[axis:]

    return v.with_return_shape(ret,
                               result_shape)

v.pushthrough_impls[(np_p.stack_p, p.cdist_p)] = push_stack_through_cdist


def push_concatenate_through_cdist_multi(expr, allow_partial=True):
    assert expr.op == np_p.concatenate_p
    assert all(child_expr.op == csp_p.cdist_multi_p for child_expr in expr.args[0])

    # TODO process this
    concat_axis = expr.kwargs.get("axis", 0)

    left = []
    right = []
    lengths = []
    axes = []
    for child_expr in expr.args[0]:
        left.append(child_expr.args[0])
        right.append(child_expr.args[1])
        lengths.append(child_expr.kwargs["lengths"])
        axes.append(child_expr.kwargs.get("axis", None))

    canonicalized_axes = [(axis if axis is not None else 0)
                          for axis in axes]
    if not all(axis == canonicalized_axes[0]
               for axis in canonicalized_axes[1:]):
        raise ValueError("Expected same axes", axes)
    axis = axes[0]

    left = v._vectorize(vnp.concatenate(left, axis=-1))
    right = v._vectorize(vnp.concatenate(right, axis=-1))

    kwargs = dict(
        lengths=np.concatenate(lengths)
    )
    if axis is not None:
        kwargs["axis"] = axis

    return cdist_multi(left, right, **kwargs)

v.pushthrough_impls[(np_p.concatenate_p, csp_p.cdist_multi_p)] = push_concatenate_through_cdist_multi



from functools import partial

import vexpr as vp
import vexpr.torch as vtorch
import vexpr.core as core
import vexpr.torch.primitives as p
import vexpr.custom.torch.primitives as cp
import vexpr.vectorization
from vexpr.vectorization import _vectorize

PRIORITIZED_OPS = set([
    p.stack_p, p.cat_p,
    p.sum_p, p.prod_p, core.operator_add_p,
    core.operator_mul_p, core.operator_truediv_p,
    core.operator_matmul_p,
    p.index_add_p, p.index_reduce_p,
    cp.index_add_into_zeros_p, cp.index_reduce_into_ones_p,
    cp.sum_multi_p, cp.prod_multi_p, cp.fast_prod_positive_p,
    cp.fast_prod_positive_multi_p,
    cp.shuffle_p,
])


def stack_vectorize(expr):
    # get unique list of ops, preserving order
    vexpr_ops = list(dict.fromkeys(v.op for v in expr.args[0]
                                   if isinstance(v, vp.Vexpr)))
    vexpr_ops = [op for op in vexpr_ops if op != core.symbol_p]
    vexpr_ops = ([op for op in vexpr_ops if op in PRIORITIZED_OPS]
                 + [op for op in vexpr_ops if op not in PRIORITIZED_OPS])

    if len(vexpr_ops) > 0:
        for allow_partial in (False, True):
            for op in vexpr_ops:
                try:
                    return vexpr.vectorization.pushthrough(expr, op,
                                                           allow_partial=allow_partial)
                except vexpr.vectorization.CannotVectorize:
                    pass

    return expr


# TODO: obviously, understand which of these impls should be the same.
def cat_vectorize(expr):
    # get unique list of ops, preserving order
    vexpr_ops = list(dict.fromkeys(v.op
                                   for v in expr.args[0]
                                   if isinstance(v, vp.Vexpr)))

    vexpr_ops = [op for op in vexpr_ops if op != core.symbol_p]
    vexpr_ops = ([op for op in vexpr_ops if op in PRIORITIZED_OPS]
                 + [op for op in vexpr_ops if op not in PRIORITIZED_OPS])

    # TODO if any child ops are a cat with the same dim, absorb their
    # children

    # TODO: for any adjacent numpy arrays, perform a cat immediately?

    if len(vexpr_ops) > 0:
        # first pass, detect any stacks and vectorize them, attempting to make them disappear.
        if p.stack_p in vexpr_ops:
            child_exprs = []
            changed = False
            for child_expr in expr.args[0]:
                if child_expr.op == p.stack_p:
                    try:
                        child_expr2 = _vectorize(child_expr)
                        # TODO find a faster way of doing this. probably make
                        # _vectorize indicate whether anything changed, perhaps
                        # always throwing CannotVectorize if no change occurs.
                        if child_expr != child_expr2:
                            child_expr = child_expr2
                            changed = True
                    except vexpr.vectorization.CannotVectorize:
                        pass
                child_exprs.append(child_expr)
            if changed:
                return cat_vectorize(vtorch.cat(child_exprs, **expr.kwargs))

        for allow_partial in (False, True):
            for op in vexpr_ops:
                try:
                    return vexpr.vectorization.pushthrough(
                        expr, op, allow_partial=allow_partial)
                except vexpr.vectorization.CannotVectorize:
                    pass

    if len(expr.args[0]) == 1:
        # If cat-ing one item, just return it.
        return expr.args[0][0]
    else:
        return expr

def reduction_vectorize(op, expr):
    assert expr.op == op
    # TODO handle dim?
    child_expr = expr.args[0]
    if not isinstance(child_expr, core.Vexpr):
        child_expr = vtorch.stack(expr.args[0])
    child_expr = _vectorize(child_expr)
    kwargs = {}
    if "dim" in expr.kwargs:
        kwargs["dim"] = expr.kwargs["dim"]
    return core.Vexpr(op, (child_expr,), kwargs)

vexpr.vectorization.vectorize_impls.update({
    p.stack_p: stack_vectorize,
    p.cat_p: cat_vectorize,
    p.sum_p: partial(reduction_vectorize, p.sum_p),
    p.prod_p: partial(reduction_vectorize, p.prod_p),
})



import torch

import vexpr as vp
import vexpr.torch as vtorch
import vexpr.torch.primitives as p
import vexpr.custom.torch as vctorch
import vexpr.vectorization as v
from vexpr.custom.torch.utils import split_and_stack_kwargs

def torch_stack_shape(initial_shape, num_elements, dim=0):
    if dim < 0:
        dim += len(initial_shape) + 1
    return initial_shape[:dim] + (num_elements,) + initial_shape[dim:]


def torch_stack_shape2(shapes, dim=0):
    assert all(shape == shapes[0] for shape in shapes)
    return torch_stack_shape(shapes[0], len(shapes), dim)


def torch_cat_shape(child_shapes, dim=0):
    if dim < 0:
        dim += len(child_shapes[0])
    return (child_shapes[0][:dim]
            + (sum(shape[dim] for shape in child_shapes),)
            + child_shapes[0][dim + 1:])


def invert_shuffle(indices):
    inverted_indices = torch.zeros_like(torch.as_tensor(indices))
    inverted_indices[indices] = torch.arange(len(indices))
    return inverted_indices


def maybe_shuffle(expr, scrambled_indices, **kwargs):
    if torch.equal(torch.as_tensor(scrambled_indices),
                   torch.arange(len(scrambled_indices))):
        # the shuffle would be a no-op
        return expr
    else:
        indices = invert_shuffle(scrambled_indices)
        return v.with_return_shape(
            vctorch.shuffle(expr, indices, **kwargs),
            v.shape(expr))


def stack_remainder_then_combine(applicable, remainder, applicable_indices,
                                 remainder_indices, **stack_kwargs):
    if len(remainder) == 0:
        return applicable

    remainder = v._vectorize(
        v.with_return_shape(
            vtorch.stack(remainder, **stack_kwargs),
            torch_stack_shape2([v.shape(r_expr)
                                for r_expr in remainder],
                               **stack_kwargs)
        )
    )

    result_shape = torch_cat_shape([v.shape(applicable), v.shape(remainder)],
                                   **stack_kwargs)

    result = v.with_return_shape(
        vtorch.cat([applicable, remainder], **stack_kwargs),
        result_shape)

    return maybe_shuffle(result, applicable_indices + remainder_indices,
                         **stack_kwargs)


def cat_remainder_then_combine(applicable, remainder, applicable_indices,
                               remainder_indices, **cat_kwargs):
    if len(remainder) == 0:
        return applicable

    remainder = v._vectorize(
        v.with_return_shape(
            vtorch.cat(remainder, **cat_kwargs),
            torch_cat_shape([v.shape(r_expr)
                             for r_expr in remainder],
                            **cat_kwargs)
        )
    )

    result_shape = torch_cat_shape([v.shape(applicable), v.shape(remainder)],
                                   **cat_kwargs)
    result = v.with_return_shape(vtorch.cat([applicable, remainder],
                                            **cat_kwargs),
                                 result_shape)

    return maybe_shuffle(result, applicable_indices + remainder_indices,
                         **cat_kwargs)


def push_stack_through_reduction(reduction_p, parallel_reduction, fill_value,
                                 expr, allow_partial=True):
    assert expr.op == p.stack_p

    exprs_to_stack = expr.args[0]
    all_reduction_operands = []
    lengths = []

    stack_axis = expr.kwargs.get("dim", 0)

    for i, child_expr in enumerate(exprs_to_stack):
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == reduction_p:

            r_axis = child_expr.kwargs.get("dim", None)
            if r_axis is None:
                raise NotImplementedError()

            r_arg0 = child_expr.args[0]
            if isinstance(r_arg0, vp.Vexpr):
                num_operands = v.shape(r_arg0)[r_axis]
            else:
                num_operands = len(r_arg0)
                # treat array_like as an implicit stack.
                r_arg0 = v.with_return_shape(vtorch.stack(r_arg0),
                                             dict(return_shape=torch_stack_shape(
                                                 v.shape(r_arg0[0]), num_operands)))

            if r_axis != stack_axis:
                r_arg0 = vtorch.moveaxis(r_arg0, r_axis, stack_axis)
                r_arg0 = v.pushthrough(r_arg0, child_expr.op)
                r_axis = stack_axis

            # Incorporate child_expr's computation into a vectorized reduction.
            all_reduction_operands.append(r_arg0)
            lengths.append(num_operands)
        else:
            # Pass child_expr through. Implement Identity as a reduction of 1
            # element.

            # TODO is this is gross to do a full vectorize here?
            child_expr = vtorch.stack([child_expr], dim=stack_axis)
            child_expr = v._vectorize(child_expr)
            all_reduction_operands.append(child_expr)
            lengths.append(1)

    all_reduction_operands = v._vectorize(vtorch.cat(all_reduction_operands,
                                                     dim=stack_axis))
    all_reduction_operands = vctorch.split_and_stack(all_reduction_operands,
                                                     **split_and_stack_kwargs(lengths),
                                                     fill_value=fill_value,
                                                     dim=stack_axis)
    result = parallel_reduction(all_reduction_operands, dim=stack_axis)

    child_shape = next(v.shape(expr) for expr in exprs_to_stack
                       if isinstance(expr, vp.Vexpr))

    return v.with_return_shape(result, torch_stack_shape(child_shape,
                                                         len(exprs_to_stack),
                                                         dim=stack_axis))



from vexpr import Vexpr, core

################################################################################
# Vexpr's vectorizor
#
# To vectorize an expression, we start from the root node and for each arg we
# move downward through the tree, attempting to convert each level into an
# operation on a single array. The process typically starts by taking an
# expression and wrapping each of its aruments with a "stack" operation, then
# attempting a "pushthrough" of that stack operation through that argument's
# descendants. As the "stack" operation pushes through each level, its effect
# changes; for example, when you pushthrough a "stack" through an array of
# "sum"s, the logic should then pushthrough a "concatenate" through the sums'
# children. This pushthrough logic is specified by for op-child_op pairs via
# pushthrough_impls.
################################################################################


class VexprWithMetadata(Vexpr):
    metadata: dict

    def __new__(cls, op, args, kwargs, metadata):
        self = super().__new__(cls, op, args, kwargs)
        self.metadata = metadata
        return self


def with_metadata(expr: Vexpr, metadata: dict):
    return VexprWithMetadata(*expr, metadata)


def with_return_shape(expr: Vexpr, return_shape: tuple):
    return with_metadata(expr, dict(return_shape=return_shape))


shape_impls = {
    int: lambda _: (),
    float: lambda _: (),
    VexprWithMetadata: lambda expr: expr.metadata["return_shape"],
}


def shape(expr):
    impl = shape_impls[type(expr)]
    return impl(expr)


def evaluate_args(args, context):
    args_with_metadata = []
    evaluated_args = []

    for arg in args:
        if isinstance(arg, Vexpr):
            arg, v = traced_call(arg, context)
            args_with_metadata.append(arg)
            evaluated_args.append(v)
        elif isinstance(arg, (list, tuple)):
            (sub_args_with_metadata,
             sub_evaluated_args) = evaluate_args(arg, context)
            args_with_metadata.append(sub_args_with_metadata)
            evaluated_args.append(sub_evaluated_args)
        else:
            evaluated_args.append(arg)
            args_with_metadata.append(arg)

    container = type(args)
    return container(args_with_metadata), container(evaluated_args)


def traced_call(expr, context):
    """
    This function implements the Vexpr interpreter.
    """
    if not isinstance(expr, Vexpr):
        raise ValueError(expr)

    if expr.op == core.symbol_p:
        result = context[expr.args[0]]
    elif expr.op == core.let_p:
        raise NotImplementedError()
    else:
        args_with_metadata, evaluated_args = evaluate_args(expr.args, context)
        expr = Vexpr(expr.op, args_with_metadata, expr.kwargs)
        impl = core.eval_impls[expr.op]
        result = impl(*evaluated_args, **expr.kwargs)

    expr = with_metadata(expr, dict(return_shape=shape(result)))

    return expr, result


# {op: f(shapes, expr)}
vectorize_impls = {}

def _vectorize(expr):
    # assert isinstance(expr, VexprWithMetadata)

    impl = vectorize_impls.get(expr.op, None)
    if impl is None:
        print("No vectorization support for", expr.op)
        return expr
    else:
        return impl(expr)


# {(op, child_op): f(expr, allow_partial)}
pushthrough_impls = {}

def pushthrough(expr, child_op, allow_partial=True):
    # assert isinstance(expr, VexprWithMetadata)

    impl = pushthrough_impls.get((expr.op, child_op), None)
    if impl is None:
        print("No vectorization support for", expr.op, child_op)
        raise CannotVectorize()

    return impl(expr, allow_partial)


################################################################################
# Vectorize implementations for operators
################################################################################

def unary_elementwise_vectorize(expr):
    return Vexpr(expr.op, (_vectorize(expr.args[0]),), expr.kwargs)

def operator_vectorize(expr):
    args = tuple(_vectorize(arg) for arg in expr.args)
    return Vexpr(expr.op, args, {})

def identity(expr): return expr

vectorize_impls.update({
    core.symbol_p: identity,
    core.operator_add_p: operator_vectorize,
    core.operator_mul_p: operator_vectorize,
    core.operator_truediv_p: operator_vectorize,
    core.operator_pow_p: operator_vectorize,
    core.operator_matmul_p: operator_vectorize,
    core.operator_neg_p: unary_elementwise_vectorize,
    core.operator_getitem_p: operator_vectorize,
})


class CannotVectorize(Exception):
    pass



################################################################################
# User interface
################################################################################


def args_with_stripped_metadata(args):
    new_args = []
    for arg in args:
        if isinstance(arg, Vexpr):
            arg = strip_metadata(arg)
        elif isinstance(arg, (list, tuple)):
            arg = args_with_stripped_metadata(arg)
        new_args.append(arg)
    return type(args)(new_args)


def strip_metadata(expr):
    return Vexpr(expr.op, args_with_stripped_metadata(expr.args), expr.kwargs)


def call_and_vectorize(vexpr_caller, i_alternate, **kwargs):
    expr, result = traced_call(vexpr_caller.vexpr, kwargs)
    vexpr_caller.vexpr = strip_metadata(_vectorize(expr))
    del vexpr_caller.alternate_calls[i_alternate]
    return result


def vectorize(f_or_expr, example_inputs=None):
    if example_inputs is not None:
        if isinstance(f_or_expr, Vexpr):
            expr, result = traced_call(f_or_expr, example_inputs)
            return strip_metadata(_vectorize(expr))
        else:
            if isinstance(f_or_expr, core.VexprCaller):
                vexpr_caller = f_or_expr.clone()
            else:
                vexpr_caller = core.make_vexpr(f_or_expr)

            expr, result = traced_call(vexpr_caller.vexpr, example_inputs)
            vexpr_caller.vexpr = strip_metadata(_vectorize(expr))
            return vexpr_caller
    else:
        # Lazily vectorize when inputs are provided
        if isinstance(f_or_expr, core.VexprCaller):
            vexpr_caller = f_or_expr.clone()
        elif isinstance(f_or_expr, Vexpr):
            vexpr_caller = core.VexprCaller(f_or_expr, [])
        else:
            vexpr_caller = core.make_vexpr(f_or_expr)

        vexpr_caller.alternate_calls.append(call_and_vectorize)
        return vexpr_caller



from functools import partial

import torch

import vexpr as vp
import vexpr.core as core
import vexpr.custom.torch as vctorch
import vexpr.custom.torch.primitives as csp_p
import vexpr.torch as vtorch
import vexpr.torch.primitives as p
import vexpr.vectorization as v
from vexpr.vectorization import _vectorize, with_metadata
from vexpr.custom.torch.utils import (
    split_and_stack_kwargs,
)
from vexpr.torch.utils import (
    torch_cat_shape,
    torch_stack_shape,
    torch_stack_shape2,
    stack_remainder_then_combine,
    cat_remainder_then_combine,
    push_stack_through_reduction,
)


def canonical_axis(axis, ndim):
    if axis is None:
        return None
    elif axis < 0:
        return axis + ndim
    else:
        return axis

def push_cat_through_getitem(expr, allow_partial=True):
    assert expr.op == p.cat_p

    # Quick hacky assumptions: every child expr is selecting from a symbol, and
    # they are the same symbol.
    # TODO: if any child is not a getitem, then push through all the getitems,
    # cat with the non-getitems, and insert a shuffle after everything
    # to correct the order.
    if not all(sub_expr.op == core.operator_getitem_p
               for sub_expr in expr.args[0]):
        raise NotImplementedError()

    select_targets = [sub_expr.args[0]
                      for sub_expr in expr.args[0]]
    select_target = select_targets[0]
    if not all(target.op == core.symbol_p for target in select_targets):
        raise NotImplementedError()
    if not all(target == select_target
               for target in select_targets):
        raise NotImplementedError()

    selections = [sub_expr.args[1]
                  for sub_expr in expr.args[0]]

    all_select_axes = []
    all_select_indices = []

    for selection in selections:
        if isinstance(selection, tuple):
            # parse selection tuple to determine which axis contains a list of
            # indices
            select_axis = None
            select_indices = None
            for i, sel in enumerate(selection):
                if isinstance(sel, slice):
                    continue
                elif sel is Ellipsis:
                    break
                else:
                    if select_axis is not None:
                        raise NotImplementedError(
                            "Can't handle indexing on multiple axes yet.")
                    select_axis = i
                    select_indices = sel

            if select_axis is None:
                # an ellipsis was found, count from the back
                for i, sel in enumerate(reversed(selection)):
                    if isinstance(sel, slice):
                        continue
                    elif sel is Ellipsis:
                        if select_axis is None:
                            raise ValueError(
                                "Found multiple ellipses, or found no indices",
                                selection)
                    else:
                        if select_axis is not None:
                            raise NotImplementedError(
                                "Can't handle indexing on multiple axes yet.")
                        select_axis = -(i + 1)
                        select_indices = sel

        else:
            select_axis = 0
            select_indices = selection

        all_select_axes.append(select_axis)
        all_select_indices.append(torch.as_tensor(select_indices))

    if not all(select_axis == all_select_axes[0] for select_axis in all_select_axes):
        raise NotImplementedError()
    select_axis = all_select_axes[0]

    axis = expr.kwargs.get("dim", 0)
    select_target_shape = v.shape(select_target)
    ndim = len(select_target_shape)
    if not canonical_axis(axis, ndim) == canonical_axis(select_axis, ndim):
        raise NotImplementedError()
    else:
        # prefer the version that the user specified
        axis = select_axis

    indices = torch.cat(all_select_indices)

    if torch.equal(indices, torch.arange(select_target_shape[axis])):
        # the vectorized getitem is selecting all elements, so it can be
        # factored out.
        return select_target

    if axis == 0:
        new_selection = indices
    elif axis < 0:
        new_selection = (Ellipsis, indices,) + (slice(None),) * ((-axis) - 1)
    else:
        new_selection = ((slice(None),) * axis) + (indices,)

    return select_target[new_selection]

v.pushthrough_impls.update({
    (p.cat_p, core.operator_getitem_p): push_cat_through_getitem,
})


def push_moveaxis_through_stack(expr, allow_partial=True):
    assert expr.op == p.moveaxis_p
    assert isinstance(expr.args[0], vp.Vexpr) and expr.args[0].op == p.stack_p

    source = expr.args[1]
    stack_expr = expr.args[0]
    stack_axis = stack_expr.kwargs.get("dim", 0)

    orig_stack_shape = v.shape(stack_expr)
    ndim = len(orig_stack_shape)
    if source < 0: source += ndim
    if stack_axis < 0: stack_axis += ndim

    if stack_axis == source:
        dest = expr.args[2]
        new_shape = list(orig_stack_shape)
        tmp = new_shape[source]
        new_shape[source] = new_shape[dest]
        new_shape[dest] = tmp
        return v.with_return_shape(vtorch.stack(stack_expr.args[0], dim=dest),
                                   tuple(new_shape))
    else:
        # don't attempt, leave the moveaxis where it is
        return expr


def push_cat_through_zeros_ones(zeros_ones_p, zeros_ones,
                                   expr, allow_partial=True):
    assert expr.op == p.cat_p

    # initial hack: assume all args are same op
    assert all(isinstance(arg, vp.Vexpr) and arg.op == zeros_ones_p
               for arg in expr.args[0])

    dim = expr.kwargs.get("dim", 0)
    child_shapes = [child_expr.args[0] for child_expr in expr.args[0]]
    result_shape = torch_cat_shape(child_shapes, dim)
    return zeros_ones(result_shape)

v.pushthrough_impls.update({
    (p.cat_p, p.zeros_p): partial(push_cat_through_zeros_ones, p.zeros_p,
                                  vtorch.zeros),
    (p.cat_p, p.ones_p): partial(push_cat_through_zeros_ones, p.ones_p,
                                 vtorch.ones),
})

def push_moveaxis_through_sum(expr, allow_partial=True):
    assert expr.op == p.moveaxis_p
    assert isinstance(expr.args[0], vp.Vexpr) and expr.args[0].op == p.sum_p

    sum_expr = expr.args[0]
    sum_arg0 = sum_expr.args[0]
    if not isinstance(sum_arg0, vp.Vexpr):
        sum_arg0 = v.with_return_shape(vtorch.stack(sum_arg0),
                                       torch_stack_shape(
                                           v.shape(sum_expr.args[0]),
                                           len(sum_arg0), dim=0))

    source = expr.args[1]
    dest = expr.args[2]
    sum_arg0 = vtorch.moveaxis(sum_arg0, source, dest)
    sum_arg0 = v.pushthrough(sum_arg0, p.stack_p)

    sum_axis = sum_expr.kwargs.get("dim", None)

    trace = list(range(len(v.shape(sum_expr))))
    trace.insert(dest, trace.pop(source))
    new_axis = trace.index(sum_axis)

    return sum(sum_arg0, dim=new_axis)

v.pushthrough_impls.update({
    (p.moveaxis_p, p.stack_p): push_moveaxis_through_stack,
    (p.moveaxis_p, p.sum_p): push_moveaxis_through_sum,
})


def push_cat_through_stack(expr, allow_partial=True):
    assert expr.op == p.cat_p

    # initial hack: assume all args are stack_p
    assert all(isinstance(arg, vp.Vexpr) and arg.op == p.stack_p
               for arg in expr.args[0])

    dim = expr.kwargs.get("dim", 0)

    # (when i remove the hack, the result is going to be
    # shuffle(cat([stack, non-stack-exprs])), or remove the outer
    # cat of there are no non-stack-exprs.)

    # todo push vectorize through children
    all_stacked_vexprs = []
    for child_expr in expr.args[0]:
        assert child_expr.kwargs.get("dim", 0) == dim
        all_stacked_vexprs.extend(child_expr.args[0])

    assert all(child_expr.kwargs == expr.args[0][0].kwargs
               for child_expr in expr.args[0][1:])
    stack_kwargs = expr.args[0][0].kwargs

    expr = v.with_return_shape(
        vtorch.stack(all_stacked_vexprs, **stack_kwargs),
        torch_stack_shape2([v.shape(d) for d in all_stacked_vexprs],
                           **stack_kwargs)
    )
    expr = v._vectorize(expr)
    return expr

def push_cat_through_cat(expr, allow_partial=True):
    assert expr.op == p.cat_p

    cat_dims = [child_expr.kwargs.get("dim", 0)
                for child_expr in expr.args[0]
                if isinstance(child_expr, vp.Vexpr)
                and child_expr.op == p.cat_p]
    assert all(cat_dim == cat_dims[0] for cat_dim in cat_dims)
    cat_dim = cat_dims[0]

    # todo push vectorize through children
    all_cat_vexprs = []
    for child_expr in expr.args[0]:
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == p.cat_p:
            all_cat_vexprs.extend(child_expr.args[0])
        else:
            all_cat_vexprs.append(child_expr)

    expr = vtorch.cat(all_cat_vexprs, dim=cat_dim)
    return v.with_return_shape(
        v._vectorize(expr),
        torch_cat_shape([v.shape(child_expr) for child_expr in all_cat_vexprs],
                        cat_dim)
    )


def push_cat_through_index_reduction(index_reduction_p, parallel_reduction,
                                     expr, allow_partial=True):
    assert expr.op == p.cat_p

    cat_dim = expr.kwargs.get("dim", 0)

    index_reduction_dims = [child_expr.args[1]
                            for child_expr in expr.args[0]
                            if isinstance(child_expr, vp.Vexpr)
                            and child_expr.op == index_reduction_p]
    assert all(dim == index_reduction_dims[0] for dim in index_reduction_dims)
    index_reduction_dim = index_reduction_dims[0]

    target = []
    indices = []
    grandchildren = []
    base = 0
    for child_expr in expr.args[0]:
        child_shape = v.shape(child_expr)
        num_results = child_shape[cat_dim]
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == index_reduction_p:
            target.append(child_expr.args[0])
            grandchildren.append(child_expr.args[3])
            indices.append(child_expr.args[2] + base)
            base += num_results
        else:
            target.append(vtorch.zeros(child_shape))
            grandchildren.append(child_expr)
            indices.append(torch.arange(base, base + num_results))
            base += num_results

    target = v._vectorize(vtorch.cat(target, dim=cat_dim))
    indices = torch.cat(indices)
    grandchildren = v._vectorize(vtorch.cat(grandchildren, dim=cat_dim))

    return v.with_return_shape(parallel_reduction(target, index_reduction_dim,
                                                  indices, grandchildren),
                               torch_cat_shape([v.shape(child_expr)
                                                for child_expr in expr.args[0]],
                                               dim=cat_dim))

def parallel_sum(target, dim, index, source):
    return vtorch.index_add(target, dim, index, source)

def parallel_prod(target, dim, index, source):
    return vtorch.index_reduce(target, dim, index, source, "prod")


def push_stack_through_unary_elementwise(op, expr, allow_partial=True):
    assert expr.op == p.stack_p

    applicable = []
    applicable_indices = []
    remainder = []
    remainder_indices = []
    for i, child_expr in enumerate(expr.args[0]):
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == op:
            applicable.append(child_expr)
            applicable_indices.append(i)
        else:
            remainder.append(child_expr)
            remainder_indices.append(i)

    grandchildren = [child_expr.args[0] for child_expr in applicable]
    grandchildren = v._vectorize(
        v.with_return_shape(vtorch.stack(grandchildren, **expr.kwargs),
                            torch_stack_shape2([v.shape(gc)
                                                for gc in grandchildren],
                                               **expr.kwargs)))

    return stack_remainder_then_combine(
        v.with_return_shape(
            vp.Vexpr(op, (grandchildren,), applicable[0].kwargs),
            v.shape(grandchildren)
        ),
        remainder,
        applicable_indices,
        remainder_indices,
        **expr.kwargs)


def push_cat_through_unary_elementwise(op, expr, allow_partial=True):
    assert expr.op == p.cat_p

    applicable = []
    applicable_indices = []
    remainder = []
    remainder_indices = []
    base = 0
    dim = expr.kwargs.get("dim", 0)
    for child_expr in expr.args[0]:
        num_indices = v.shape(child_expr)[dim]
        result_indices = list(range(base, base + num_indices))
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == op:
            applicable.append(child_expr)
            applicable_indices += result_indices
        else:
            if not allow_partial:
                raise v.CannotVectorize()
            remainder.append(child_expr)
            remainder_indices += result_indices
        base += num_indices

    grandchildren = [child_expr.args[0] for child_expr in applicable]
    grandchildren = v._vectorize(
        v.with_return_shape(vtorch.cat(grandchildren, **expr.kwargs),
                            torch_cat_shape([v.shape(gc)
                                             for gc in grandchildren],
                                            **expr.kwargs)))

    return cat_remainder_then_combine(
        v.with_return_shape(
            vp.Vexpr(op, (grandchildren,), applicable[0].kwargs),
            v.shape(grandchildren)
        ),
        remainder,
        applicable_indices,
        remainder_indices,
        **expr.kwargs)


def push_cat_through_truediv(expr, allow_partial=True):
    assert expr.op == p.cat_p

    # initial hack: assume all args are truediv
    #
    # TODO: add similar treatment of "*_through_mul", dividing by 1.0. It's
    # easier than mul, since it's unambiguous; we can always put 1.0 in the
    # denominator.
    assert all(isinstance(arg, vp.Vexpr) and arg.op == core.operator_truediv_p
               for arg in expr.args[0])

    num = []
    den = []
    for child_expr in expr.args[0]:
        num.append(child_expr.args[0])
        den.append(child_expr.args[1])

    axis = expr.kwargs.get("dim", 0)
    num = v._vectorize(vtorch.cat(num, dim=axis))
    den = v._vectorize(vtorch.cat(den, dim=axis))

    return num / den

def push_stack_through_mul(expr, allow_partial=True):
    assert expr.op == p.stack_p

    left = []
    right = []
    for child_expr in expr.args[0]:
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == core.operator_mul_p:
            left.append(child_expr.args[0])
            right.append(child_expr.args[1])
        else:
            # TODO: how to choose side? for example, some users do w * matrix,
            # while others do matrix * w. Here I want to append the 1 to the w.
            # It seems the way to do this is to check which symbols each
            # child_expr is derived from, then when we're doing an identity, use
            # the symbols *this* child_expr is derived from. if they aren't
            # derived from the same symbols, then there's no benefit to choosing
            # correctly.
            #
            # TODO: what if we're multiplying matrices with matrices? sometimes
            # we need to instead append, e.g., torch.ones((1, 1, 1)) when we're
            # weighting matrices, because these "ones" need to be able to
            # concatenate with other actual tensors of weights.
            left.append(1.0)
            right.append(child_expr)

    kwargs = {}
    if "dim" in expr.kwargs:
        kwargs["dim"] = expr.kwargs["dim"]

    # decide whether ones need to be inserted into the shape to get proper
    # broadcasting
    left_shapes = [v.shape(child_expr) for child_expr in left]
    left_shape_before_stack = left_shapes[0]
    assert all(left_shape_before_stack == shape for shape in left_shapes)
    right_shapes = [v.shape(child_expr) for child_expr in right]
    right_shape_before_stack = right_shapes[0]
    assert all(right_shape_before_stack == shape for shape in right_shapes)
    discrepancy = len(right_shape_before_stack) - len(left_shape_before_stack)
    if discrepancy == 0:
        implicit_left_shape = left_shape_before_stack
        implicit_right_shape = right_shape_before_stack
    elif discrepancy > 0:
        implicit_left_shape = left_shape_before_stack + (1,) * discrepancy
        implicit_right_shape = right_shape_before_stack
    else:
        implicit_left_shape = left_shape_before_stack
        implicit_right_shape = right_shape_before_stack + (1,) * -discrepancy

    axis = expr.kwargs.get("dim", 0)
    actual_left_shape = torch_stack_shape(left_shape_before_stack,
                                          len(left), axis)
    actual_right_shape = torch_stack_shape(right_shape_before_stack,
                                           len(right), axis)
    implicit_left_shape = torch_stack_shape(implicit_left_shape,
                                            len(left), axis)
    implicit_right_shape = torch_stack_shape(implicit_right_shape,
                                             len(right), axis)

    implicit_neg_axis = (axis
                         if axis < 0
                         else -len(implicit_left_shape) + axis)

    left = v._vectorize(v.with_return_shape(vtorch.stack(left, **kwargs),
                                            actual_left_shape))
    if implicit_left_shape[implicit_neg_axis:] \
       != actual_left_shape[implicit_neg_axis:]:
        shape = implicit_left_shape[implicit_neg_axis:]
        left = v.with_return_shape(vtorch.reshape(left, shape),
                                   shape)

    right = v._vectorize(v.with_return_shape(vtorch.stack(right, **kwargs),
                                             actual_right_shape))
    if implicit_right_shape[implicit_neg_axis:] \
       != actual_right_shape[implicit_neg_axis:]:
        shape = implicit_right_shape[implicit_neg_axis:]
        right = v.with_return_shape(vtorch.reshape(right, shape),
                                    shape)

    return v.with_return_shape(left * right,
                               v.shape(right))

def push_cat_through_mul(expr, allow_partial=True):
    assert expr.op == p.cat_p

    dim = expr.kwargs.get("dim", 0)

    child_shapes = [v.shape(child_expr) for child_expr in expr.args[0]]
    ndim = len(child_shapes[0])
    ndim_identity_broadcast = ndim - canonical_axis(dim, ndim) - 1

    left = []
    right = []
    base = 0
    actual_indices = []
    identity = False
    for child_expr in expr.args[0]:
        n = v.shape(child_expr)[dim]
        if isinstance(child_expr, vp.Vexpr) and child_expr.op == core.operator_mul_p:
            left.append(child_expr.args[0])
            right.append(child_expr.args[1])
            actual_indices += list(range(base, base + n))
        else:
            # TODO: this is hardcoded to expect identity to occur on the left
            right.append(child_expr)
            identity = True
        base += n
    total_n = base

    left = v._vectorize(vtorch.cat(left, dim=dim))
    if identity:
        ones_shape = (total_n,) + (1,) * ndim_identity_broadcast
        left = v.with_return_shape(
            vtorch.scatter(
                v.with_return_shape(vtorch.ones(ones_shape),
                                    ones_shape),
                dim,
                torch.tensor(actual_indices),
                left),
            ones_shape)

    right = v._vectorize(vtorch.cat(right, dim=dim))

    return v.with_return_shape(left * right, torch_cat_shape(child_shapes,
                                                             dim=dim))

def push_cat_through_scatter(expr, allow_partial):
    assert expr.op == p.cat_p

    into = []
    indices = []
    sources = []

    dim = expr.kwargs.get("dim", 0)

    base = 0
    for child_expr in expr.args[0]:
        if not isinstance(child_expr, vp.Vexpr) or child_expr.op != p.scatter_p:
            raise NotImplementedError()

        assert child_expr.args[1] == dim

        into.append(child_expr.args[0])
        indices.append(base + child_expr.args[2])
        sources.append(child_expr.args[3])

        base += v.shape(child_expr)[dim]

    into_shape = torch_cat_shape([v.shape(into_expr) for into_expr in into],
                                 dim=dim)

    into = v.with_return_shape(v._vectorize(vtorch.cat(into, dim=dim)),
                               into_shape)
    indices = torch.cat(indices)
    sources = v._vectorize(vtorch.cat(sources, dim=dim))
    return v.with_return_shape(vtorch.scatter(into, dim, indices, sources),
                               into_shape)

v.pushthrough_impls.update({
    (p.cat_p, p.stack_p): push_cat_through_stack,
    (p.cat_p, p.cat_p): push_cat_through_cat,
    (p.stack_p, p.sum_p): partial(push_stack_through_reduction, p.sum_p,
                                  vctorch.sum_multi, 0.),
    (p.stack_p, p.prod_p): partial(push_stack_through_reduction, p.prod_p,
                                   vctorch.prod_multi, 1.),
    (p.stack_p, csp_p.fast_prod_positive_p): partial(
        push_stack_through_reduction, csp_p.fast_prod_positive_p,
        vctorch.fast_prod_positive_multi, 1.),
    (p.cat_p, p.index_add_p): partial(push_cat_through_index_reduction,
                                      p.index_add_p,
                                      parallel_sum),
    # TODO this is hardcoded to prod, but index reduce might use e.g. mean
    (p.cat_p, p.index_reduce_p): partial(push_cat_through_index_reduction,
                                         p.index_reduce_p,
                                         parallel_prod),
    (p.stack_p, p.exp_p): partial(
        push_stack_through_unary_elementwise, p.exp_p),
    (p.cat_p, p.exp_p): partial(
        push_cat_through_unary_elementwise, p.exp_p),
    (p.cat_p, core.operator_truediv_p): push_cat_through_truediv,
    (p.stack_p, core.operator_mul_p): push_stack_through_mul,
    (p.cat_p, core.operator_mul_p): push_cat_through_mul,
    (p.stack_p, core.operator_neg_p): partial(
        push_stack_through_unary_elementwise, core.operator_neg_p),
    (p.cat_p, core.operator_neg_p): partial(
        push_cat_through_unary_elementwise, core.operator_neg_p),
    (p.cat_p, p.scatter_p): push_cat_through_scatter,
})


def push_stack_through_cdist(expr, allow_partial=True):
    assert expr.op == p.stack_p

    metric = next(child_expr.kwargs.get("p", 2)
                  for child_expr in expr.args[0]
                  if child_expr.op == p.cdist_p)

    # TODO process this
    stack_axis = expr.kwargs.get("dim", 0)

    left = []
    right = []
    lengths = []
    child_matrix_shapes = []

    applicable_indices = []
    remainder = []
    remainder_indices = []

    for i, child_expr in enumerate(expr.args[0]):
        if (child_expr.op == p.cdist_p
            and child_expr.kwargs.get("p", 2) == metric):
            applicable_indices.append(i)
            left.append(child_expr.args[0])
            right.append(child_expr.args[1])
            lengths.append(v.shape(child_expr.args[0])[-1])
            child_matrix_shapes.append(v.shape(child_expr))
        else:
            remainder.append(child_expr)
            remainder_indices.append(i)

    left = v._vectorize(vtorch.cat(left, dim=-1))
    right = v._vectorize(vtorch.cat(right, dim=-1))

    expansion_kwargs = split_and_stack_kwargs(lengths)
    left = vctorch.split_and_stack(left, **expansion_kwargs, dim=-1)
    right = vctorch.split_and_stack(right, **expansion_kwargs, dim=-1)

    kwargs = dict(p=metric)
    kwargs["p"] = metric
    if "dim" in expr.kwargs:
        kwargs["dim"] = expr.kwargs["dim"]

    # Compute shape of result
    child_matrix_shape = child_matrix_shapes[0]
    assert all(shape == child_matrix_shape for shape in child_matrix_shapes[1:])

    axis = expr.kwargs.get("dim", 0)
    # use axis to determine result shape after stack
    result_shape = child_matrix_shape
    if axis < 0:
        axis += len(result_shape) + 1
    result_shape = result_shape[:axis] + (len(lengths),) + result_shape[axis:]
    applicable =  v.with_return_shape(
        vctorch.cdist_multi(left, right, **kwargs),
        result_shape)

    return stack_remainder_then_combine(
        v.with_return_shape(
            vctorch.cdist_multi(left, right, **kwargs),
            result_shape),
        remainder,
        applicable_indices,
        remainder_indices,
        **expr.kwargs)

v.pushthrough_impls[(p.stack_p, p.cdist_p)] = push_stack_through_cdist



from vexpr.core import _p_and_constructor

# Array construction
zeros_p, zeros = _p_and_constructor("torch.zeros")
ones_p, ones = _p_and_constructor("torch.ones")
# Array manipulation
stack_p, stack = _p_and_constructor("torch.stack")
cat_p, cat = _p_and_constructor("torch.cat")
scatter_p, scatter = _p_and_constructor("torch.scatter")
moveaxis_p, moveaxis = _p_and_constructor("torch.moveaxis")
reshape_p, reshape = _p_and_constructor("torch.reshape")
# Unary operators
exp_p, exp = _p_and_constructor("torch.exp")
log_p, log = _p_and_constructor("torch.log")
# Reduction operators
sum_p, sum = _p_and_constructor("torch.sum")
prod_p, prod = _p_and_constructor("torch.prod")
# Indexed reductions
index_add_p, index_add = _p_and_constructor("torch.index_add")
index_reduce_p, index_reduce = _p_and_constructor("torch.index_reduce")

cdist_p, cdist = _p_and_constructor("torch.cdist")

__all__ = [
    "zeros",
    "ones",
    "stack",
    "cat",
    "scatter",
    "moveaxis",
    "reshape",
    "exp",
    "log",
    "sum",
    "prod",
    "cdist",
    "index_add",
    "index_reduce",
]



from setuptools import find_packages, setup

setup(name="vexpr",
      version="0.1.0",
      description="Represent code expressions as data structures, then transform them",
      author="Marcus Lewis",
      url="https://vexpr.org/",
      packages=find_packages(),
)



from vexpr import core
from .primitives import shuffle_p

def shuffle_impl(arr, indices, axis=0):
    selection = [slice(None)] * arr.ndim
    selection[axis] = indices
    return arr[selection]

core.eval_impls[shuffle_p] = shuffle_impl


def matern_impl(x, nu=2.5):
    pass


import torch

def split_and_stack_kwargs(lengths):
    max_length = max(lengths)
    expanded_length = max_length * len(lengths)

    expanded_indices = []
    base = 0
    for length in lengths:
        expanded_indices.append(torch.arange(base, base + length))
        base += max_length
    expanded_indices = torch.cat(expanded_indices)

    return dict(
        lengths=lengths,
        expanded_length=expanded_length,
        expanded_indices=expanded_indices,
        max_length=max_length,
    )



"""
Move to README or blog post:
JAX uses a tracer-based approach for defining Jaxprs. Vexprs, on the other hand,
are built up directly by the user.
"""

import inspect
from functools import partial
from pprint import pformat
from typing import NamedTuple


class Primitive(NamedTuple):
    name: str


class Vexpr(NamedTuple):
    op: Primitive
    args: tuple
    kwargs: dict

    def __call__(self, *args, **kwargs):
        return operator_call(self, *args, **kwargs)

    def __add__(self, other):
        return operator_add(self, other)

    def __mul__(self, other):
        return operator_mul(self, other)

    def __rmul__(self, other):
        return operator_mul(other, self)

    def __truediv__(self, other):
        return operator_truediv(self, other)

    def __pow__(self, other):
        return operator_pow(self, other)

    def __matmul__(self, other):
        return operator_matmul(self, other)

    def __neg__(self):
        return operator_neg(self)

    def __getitem__(self, index):
        return operator_getitem(self, index)

    def __repr__(self):
        child_lines = []

        if len(self.args) == 1:
            child_lines += pformat(self.args[0]).split('\n')
        else:
            for arg in self.args:
                operand_str = pformat(arg) + ","
                child_lines += operand_str.split('\n')

        extra_lines = ([", ".join([f"{k}={repr(v)}"
                                   for k, v in self.kwargs.items()])]
                       if len(self.kwargs) > 0
                       else [])

        lines = child_lines + extra_lines

        main_str = self.op.name + '('
        if lines:
            if len(lines) == 1:
                main_str += lines[0]
            else:
                newline_space = "\n  "
                main_str += newline_space + newline_space.join(lines) + '\n'

        main_str += ')'
        return main_str


################################################################################
# Special primitives. The Vexpr interpreter is aware of these.
################################################################################


symbol_p = Primitive("symbol")
symbol = lambda name: Vexpr(symbol_p, (name,), {})
let_p = Primitive("let")
let = lambda bindings, expr: Vexpr(let_p, (bindings, expr), {})


################################################################################
# The Vexpr interpreter
################################################################################

eval_impls = {}
def call(expr, context):
    """
    This function implements the Vexpr interpreter.
    """
    if not isinstance(expr, Vexpr):
        raise ValueError(expr)

    if expr.op == symbol_p:
        result = context[expr.args[0]]
    elif expr.op == let_p:
        context2 = dict(context)
        for symbol, v in expr.args[0]:
            name = (symbol if isinstance(symbol, str) else symbol.args[0])
            context2[name] = call(v, context2)
        result = call(expr.args[1], context2)
    else:
        impl = eval_impls[expr.op]
        args = evaluate_args(expr.args, context)
        result = impl(*args, **expr.kwargs)

    return result


def evaluate_args(args, context, call_fn=call):
    container = type(args)
    return container((call_fn(arg, context)
                      if isinstance(arg, Vexpr)
                      else (evaluate_args(arg, context, call_fn)
                            if isinstance(arg, (list, tuple))
                            else arg))
                     for arg in args)


################################################################################
# Built-in primitives
################################################################################

def _p_and_constructor(name):
    p = Primitive(name)
    def construct_vexpr(*args, **kwargs):
        return Vexpr(p, args, kwargs)
    return p, construct_vexpr

# Python language primitives
operator_call_p, operator_call = _p_and_constructor("operator.call")
operator_add_p, operator_add = _p_and_constructor("operator.add")
operator_mul_p, operator_mul = _p_and_constructor("operator.mul")
operator_truediv_p, operator_truediv = _p_and_constructor("operator.truediv")
operator_pow_p, operator_pow = _p_and_constructor("operator.pow")
operator_matmul_p, operator_matmul = _p_and_constructor("operator.matmul")
operator_neg_p, operator_neg = _p_and_constructor("operator.neg")
operator_getitem_p, operator_getitem = _p_and_constructor("operator.getitem")

constant_p = Primitive("constant")
constant = lambda value: Vexpr(constant_p, (value,), {})

################################################################################
# Default Impls
################################################################################

import operator


eval_impls.update({
    operator_add_p: operator.add,
    operator_mul_p: operator.mul,
    operator_truediv_p: operator.truediv,
    operator_pow_p: operator.pow,
    operator_matmul_p: operator.matmul,
    operator_neg_p: operator.neg,
    operator_getitem_p: operator.getitem,
    operator_call_p: (operator.call  # This exists in Python 3.11+
                      if hasattr(operator, "call")
                      else lambda f, *args, **kwargs: f(*args, **kwargs)),
})


################################################################################
# Vexprs partial evaluation
################################################################################


# this is essentially an alternate interpreter which only evaluates an operator
# if its children can be evaluated, given the provided inputs.
def partial_eval_(expr, context):
    if not isinstance(expr, Vexpr):
        raise ValueError(expr)

    if expr.op == symbol_p:
        name = expr.args[0]
        if name in context:
            return context[expr.args[0]]
        else:
            return expr
    elif expr.op == let_p:
        context2 = dict(context)
        for symbol, v in expr.args[0]:
            name = (symbol if isinstance(symbol, str) else symbol.args[0])
            v = partial_eval_(v, context2)
            if not isinstance(v, Vexpr):
                context2[name] = v
        return partial_eval_(expr.args[1], context2)
    else:
        impl = eval_impls[expr.op]
        args = evaluate_args(expr.args, context, partial_eval_)

        ready = True
        for arg in args:
            if isinstance(arg, Vexpr):
                ready = False
            elif isinstance(arg, (list, tuple)):
                for subarg in arg:
                    if isinstance(subarg, Vexpr):
                        ready = False

        if ready:
            return impl(*args, **expr.kwargs)
        else:
            return Vexpr(expr.op, args, expr.kwargs)



################################################################################
# User interface
################################################################################

class VexprCaller:
    """
    This class enables calling a Vexpr with ordered args. It also provides a way
    of queueing alternate __call__ behavior.
    """
    def __init__(self, vexpr, arg_names):
        self.vexpr = vexpr
        self.arg_names = arg_names
        self.alternate_calls = []

    def clone(self):
        ret = VexprCaller(self.vexpr, self.arg_names)
        ret.alternate_calls = self.alternate_calls
        return ret

    def __call__(self, *args, **kwargs):
        if len(args) > 0:
            kwargs.update(dict(zip(self.arg_names, args)))

        if len(self.alternate_calls) > 0:
            call_fn = self.alternate_calls[0]
            return call_fn(self, 0, **kwargs)

        return eval(self.vexpr, kwargs)


def make_vexpr(f):
    arg_names = inspect.getfullargspec(f).args  # TODO allow kwargs?
    symbols = {name: symbol(name) for name in arg_names}
    # trace the function by calling it with symbols
    expression = f(*symbols.values())
    return VexprCaller(expression, arg_names)


def eval(expr, inputs, allow_partial=False):
    if allow_partial:
        return partial_eval_(expr, inputs)
    else:
        return call(expr, inputs)


def partial_eval(f, inputs):
    """
    If all symbols are specified, the Vexpr is fully evaluated.
    """
    if isinstance(f, VexprCaller):
        f = f.clone()
        f.vexpr = partial_eval_(f.vexpr, inputs)
        for name in inputs.keys():
            # Mimic behavior of functools.partial. Any arg after the first kwarg
            # can now only be specified as a kwarg.
            try:
                i = f.arg_names.index(name)
                del f.arg_names[i:]
            except ValueError:
                pass
        return f
    elif isinstance(f, Vexpr):
        return partial_eval_(f, inputs)
    else:
        raise ValueError(f)
